@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -34,6 +36,10 @@
 
 #include <stdio.h>
 
+#define WS_REDZONE_END		'\x15'
+
+static const uintptr_t snap_overflowed = (uintptr_t)&snap_overflowed;
+
 void
 WS_Assert(const struct ws *ws)
 {
@@ -56,7 +62,7 @@ WS_Assert(const struct ws *ws)
 		assert(ws->r <= ws->e);
 		assert(PAOK(ws->r));
 	}
-	assert(*ws->e == 0x15);
+	assert(*ws->e == WS_REDZONE_END);
 }
 
 int
@@ -92,6 +98,7 @@ WS_Assert_Allocated(const struct ws *ws, const void *ptr, ssize_t len)
 void
 WS_Init(struct ws *ws, const char *id, void *space, unsigned len)
 {
+	unsigned l;
 
 	DSL(DBG_WORKSPACE, 0,
 	    "WS_Init(%p, \"%s\", %p, %u)", ws, id, space, len);
@@ -99,16 +106,25 @@ WS_Init(struct ws *ws, const char *id, void *space, unsigned len)
 	INIT_OBJ(ws, WS_MAGIC);
 	ws->s = space;
 	assert(PAOK(space));
-	len = PRNDDN(len - 1);
-	ws->e = ws->s + len;
-	*ws->e = 0x15;
+	l = PRNDDN(len - 1);
+	ws->e = ws->s + l;
+	memset(ws->e, WS_REDZONE_END, len - l);
 	ws->f = ws->s;
-	assert(id[0] & 0x20);
-	assert(strlen(id) < sizeof ws->id);
-	strcpy(ws->id, id);
+	assert(id[0] & 0x20);		// cheesy islower()
+	bstrcpy(ws->id, id);
 	WS_Assert(ws);
 }
 
+void
+WS_Id(const struct ws *ws, char *id)
+{
+
+	WS_Assert(ws);
+	AN(id);
+	memcpy(id, ws->id, WS_ID_SIZE);
+	id[0] |= 0x20;			// cheesy tolower()
+}
+
 void
 WS_MarkOverflow(struct ws *ws)
 {
@@ -126,7 +142,12 @@ ws_ClearOverflow(struct ws *ws)
 }
 
 /*
- * Reset a WS to start or a given pointer, likely from WS_Snapshot
+ * Reset a WS to a cookie from WS_Snapshot
+ *
+ * for use by any code using cache.h
+ *
+ * does not reset the overflow bit and asserts that, if WS_Snapshot had found
+ * the workspace overflown, the marker is intact
  */
 
 void
@@ -135,20 +156,39 @@ WS_Reset(struct ws *ws, uintptr_t pp)
 	char *p;
 
 	WS_Assert(ws);
+	AN(pp);
+	if (pp == snap_overflowed) {
+		DSL(DBG_WORKSPACE, 0, "WS_Reset(%p, overflowed)", ws);
+		AN(WS_Overflowed(ws));
+		return;
+	}
 	p = (char *)pp;
 	DSL(DBG_WORKSPACE, 0, "WS_Reset(%p, %p)", ws, p);
 	assert(ws->r == NULL);
-	if (p == NULL)
-		ws->f = ws->s;
-	else {
-		assert(p >= ws->s);
-		assert(p <= ws->e);
-		ws->f = p;
-	}
-	ws_ClearOverflow(ws);
+	assert(p >= ws->s);
+	assert(p <= ws->e);
+	ws->f = p;
 	WS_Assert(ws);
 }
 
+/*
+ * Reset the WS to a cookie or its start and clears any overflow
+ *
+ * for varnishd internal use only
+ */
+
+void
+WS_Rollback(struct ws *ws, uintptr_t pp)
+{
+	WS_Assert(ws);
+
+	if (pp == 0)
+		pp = (uintptr_t)ws->s;
+
+	ws_ClearOverflow(ws);
+	WS_Reset(ws, pp);
+}
+
 void *
 WS_Alloc(struct ws *ws, unsigned bytes)
 {
@@ -195,14 +235,14 @@ WS_Copy(struct ws *ws, const void *str, int len)
 	return (r);
 }
 
-void *
+const char *
 WS_Printf(struct ws *ws, const char *fmt, ...)
 {
 	unsigned u, v;
 	va_list ap;
 	char *p;
 
-	u = WS_Reserve(ws, 0);
+	u = WS_ReserveAll(ws);
 	p = ws->f;
 	va_start(ap, fmt);
 	v = vsnprintf(p, u, fmt, ap);
@@ -223,28 +263,56 @@ WS_Snapshot(struct ws *ws)
 
 	WS_Assert(ws);
 	assert(ws->r == NULL);
+	if (WS_Overflowed(ws)) {
+		DSL(DBG_WORKSPACE, 0, "WS_Snapshot(%p) = overflowed", ws);
+		return (snap_overflowed);
+	}
 	DSL(DBG_WORKSPACE, 0, "WS_Snapshot(%p) = %p", ws, ws->f);
-	return (ws->f == ws->s ? 0 : (uintptr_t)ws->f);
+	return ((uintptr_t)ws->f);
 }
 
+/*
+ * WS_Release() must be called in all cases
+ */
 unsigned
-WS_Reserve(struct ws *ws, unsigned bytes)
+WS_ReserveAll(struct ws *ws)
+{
+	unsigned b;
+
+	WS_Assert(ws);
+	assert(ws->r == NULL);
+
+	ws->r = ws->e;
+	b = pdiff(ws->f, ws->r);
+
+	WS_Assert(ws);
+	DSL(DBG_WORKSPACE, 0, "WS_ReserveAll(%p) = %u", ws, b);
+
+	return (b);
+}
+
+/*
+ * WS_Release() must be called for retval > 0 only
+ */
+unsigned
+WS_ReserveSize(struct ws *ws, unsigned bytes)
 {
 	unsigned b2;
 
 	WS_Assert(ws);
 	assert(ws->r == NULL);
+	assert(bytes > 0);
 
 	b2 = PRNDDN(ws->e - ws->f);
-	if (bytes != 0 && bytes < b2)
+	if (bytes < b2)
 		b2 = PRNDUP(bytes);
 
-	if (ws->f + b2 > ws->e) {
+	if (bytes > b2) {
 		WS_MarkOverflow(ws);
 		return (0);
 	}
 	ws->r = ws->f + b2;
-	DSL(DBG_WORKSPACE, 0, "WS_Reserve(%p, %u/%u) = %u",
+	DSL(DBG_WORKSPACE, 0, "WS_ReserveSize(%p, %u/%u) = %u",
 	    ws, b2, bytes, pdiff(ws->f, ws->r));
 	WS_Assert(ws);
 	return (pdiff(ws->f, ws->r));
@@ -253,11 +321,7 @@ WS_Reserve(struct ws *ws, unsigned bytes)
 unsigned
 WS_ReserveLumps(struct ws *ws, size_t sz)
 {
-	unsigned u;
-
-	u = WS_Reserve(ws, 0);
-	u /= sz;
-	return (u);
+	return (WS_ReserveAll(ws) / sz);
 }
 
 void
@@ -293,7 +357,104 @@ WS_Overflowed(const struct ws *ws)
 	CHECK_OBJ_NOTNULL(ws, WS_MAGIC);
 	AN(ws->id[0]);
 
-	if (ws->id[0] & 0x20)
+	if (ws->id[0] & 0x20)		// cheesy islower()
 		return (0);
 	return (1);
 }
+
+void *
+WS_AtOffset(const struct ws *ws, unsigned off, unsigned len)
+{
+	char *ptr;
+
+	WS_Assert(ws);
+	ptr = ws->s + off;
+	WS_Assert_Allocated(ws, ptr, len);
+	return (ptr);
+}
+
+/*---------------------------------------------------------------------
+ * Build a VSB on a workspace.
+ * Usage pattern:
+ *
+ *	struct vsb vsb[1];
+ *	char *p;
+ *
+ *	WS_VSB_new(vsb, ctx->ws);
+ *	VSB_printf(vsb, "blablabla");
+ *	p = WS_VSB_finish(vsb, ctx->ws, NULL);
+ *	if (p == NULL)
+ *		return (FAILURE);
+ */
+
+void
+WS_VSB_new(struct vsb *vsb, struct ws *ws)
+{
+	unsigned u;
+	static char bogus[2];	// Smallest possible vsb
+
+	AN(vsb);
+	WS_Assert(ws);
+	u = WS_ReserveAll(ws);
+	if (WS_Overflowed(ws) || u < 2)
+		AN(VSB_init(vsb, bogus, sizeof bogus));
+	else
+		AN(VSB_init(vsb, WS_Reservation(ws), u));
+}
+
+char *
+WS_VSB_finish(struct vsb *vsb, struct ws *ws, size_t *szp)
+{
+	char *p;
+
+	AN(vsb);
+	WS_Assert(ws);
+	if (!VSB_finish(vsb)) {
+		p = VSB_data(vsb);
+		if (p == ws->f) {
+			WS_Release(ws, VSB_len(vsb) + 1);
+			if (szp != NULL)
+				*szp = VSB_len(vsb);
+			VSB_fini(vsb);
+			return (p);
+		}
+	}
+	WS_MarkOverflow(ws);
+	VSB_fini(vsb);
+	WS_Release(ws, 0);
+	if (szp)
+		*szp = 0;
+	return (NULL);
+}
+
+/*--------------------------------------------------------------------*/
+
+void
+WS_Panic(const struct ws *ws, struct vsb *vsb)
+{
+
+	VSB_printf(vsb, "ws = %p {\n", ws);
+	if (PAN_already(vsb, ws))
+		return;
+	VSB_indent(vsb, 2);
+	PAN_CheckMagic(vsb, ws, WS_MAGIC);
+	if (ws->id[0] != '\0' && (!(ws->id[0] & 0x20)))	// cheesy islower()
+		VSB_cat(vsb, "OVERFLOWED ");
+	VSB_printf(vsb, "id = \"%s\",\n", ws->id);
+	VSB_printf(vsb, "{s, f, r, e} = {%p", ws->s);
+	if (ws->f >= ws->s)
+		VSB_printf(vsb, ", +%ld", (long) (ws->f - ws->s));
+	else
+		VSB_printf(vsb, ", %p", ws->f);
+	if (ws->r >= ws->s)
+		VSB_printf(vsb, ", +%ld", (long) (ws->r - ws->s));
+	else
+		VSB_printf(vsb, ", %p", ws->r);
+	if (ws->e >= ws->s)
+		VSB_printf(vsb, ", +%ld", (long) (ws->e - ws->s));
+	else
+		VSB_printf(vsb, ", %p", ws->e);
+	VSB_cat(vsb, "},\n");
+	VSB_indent(vsb, -2);
+	VSB_cat(vsb, "},\n");
+}
