@@ -5,6 +5,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -53,14 +55,14 @@ struct v1l {
 	unsigned		magic;
 #define V1L_MAGIC		0x2f2142e5
 	int			*wfd;
-	unsigned		werr;	/* valid after V1L_Flush() */
+	enum sess_close		werr;	/* valid after V1L_Flush() */
 	struct iovec		*iov;
 	unsigned		siov;
 	unsigned		niov;
 	ssize_t			liov;
 	ssize_t			cliov;
 	unsigned		ciov;	/* Chunked header marker */
-	double			t0;
+	vtim_real		deadline;
 	struct vsl_log		*vsl;
 	ssize_t			cnt;	/* Flushed byte count */
 	struct ws		*ws;
@@ -74,7 +76,7 @@ struct v1l {
 
 void
 V1L_Open(struct worker *wrk, struct ws *ws, int *fd, struct vsl_log *vsl,
-    double t0, unsigned niov)
+    vtim_real deadline, unsigned niov)
 {
 	struct v1l *v1l;
 	unsigned u;
@@ -110,11 +112,11 @@ V1L_Open(struct worker *wrk, struct ws *ws, int *fd, struct vsl_log *vsl,
 		u = IOV_MAX;
 	if (niov != 0 && u > niov)
 		u = niov;
-	v1l->iov = (void*)ws->f;
+	v1l->iov = WS_Reservation(ws);
 	v1l->siov = u;
 	v1l->ciov = u;
 	v1l->wfd = fd;
-	v1l->t0 = t0;
+	v1l->deadline = deadline;
 	v1l->vsl = vsl;
 	wrk->v1l = v1l;
 
@@ -122,27 +124,28 @@ V1L_Open(struct worker *wrk, struct ws *ws, int *fd, struct vsl_log *vsl,
 		WS_Release(ws, u * sizeof(struct iovec));
 }
 
-unsigned
+enum sess_close
 V1L_Close(struct worker *wrk, uint64_t *cnt)
 {
 	struct v1l *v1l;
-	unsigned u;
+	enum sess_close sc;
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	AN(cnt);
-	u = V1L_Flush(wrk);
+	sc = V1L_Flush(wrk);
 	v1l = wrk->v1l;
 	wrk->v1l = NULL;
 	CHECK_OBJ_NOTNULL(v1l, V1L_MAGIC);
 	*cnt = v1l->cnt;
-	if (v1l->ws->r)
+	if (WS_IsReserved(v1l->ws))
 		WS_Release(v1l->ws, 0);
-	WS_Reset(v1l->ws, v1l->res);
-	return (u);
+	WS_Rollback(v1l->ws, v1l->res);
+	ZERO_OBJ(v1l, sizeof *v1l);
+	return (sc);
 }
 
 static void
-v1l_prune(struct v1l *v1l, ssize_t bytes)
+v1l_prune(struct v1l *v1l, size_t bytes)
 {
 	ssize_t used = 0;
 	ssize_t j, used_here;
@@ -165,7 +168,7 @@ v1l_prune(struct v1l *v1l, ssize_t bytes)
 	AZ(v1l->liov);
 }
 
-unsigned
+enum sess_close
 V1L_Flush(const struct worker *wrk)
 {
 	ssize_t i;
@@ -179,7 +182,7 @@ V1L_Flush(const struct worker *wrk)
 
 	assert(v1l->niov <= v1l->siov);
 
-	if (*v1l->wfd >= 0 && v1l->liov > 0 && v1l->werr == 0) {
+	if (*v1l->wfd >= 0 && v1l->liov > 0 && v1l->werr == SC_NULL) {
 		if (v1l->ciov < v1l->siov && v1l->cliov > 0) {
 			/* Add chunk head & tail */
 			bprintf(cbuf, "00%zx\r\n", v1l->cliov);
@@ -197,19 +200,9 @@ V1L_Flush(const struct worker *wrk)
 			v1l->iov[v1l->ciov].iov_len = 0;
 		}
 
-		i = writev(*v1l->wfd, v1l->iov, v1l->niov);
-		if (i > 0)
-			v1l->cnt += i;
-		while (i != v1l->liov && i > 0) {
-			/* Remove sent data from start of I/O vector,
-			 * then retry; we hit a timeout, but some data
-			 * was sent.
-			 *
-			 * XXX: Add a "minimum sent data per timeout
-			 * counter to prevent slowloris attacks
-			*/
-
-			if (VTIM_real() - v1l->t0 > cache_param->send_timeout) {
+		i = 0;
+		do {
+			if (VTIM_real() > v1l->deadline) {
 				VSLb(v1l->vsl, SLT_Debug,
 				    "Hit total send timeout, "
 				    "wrote = %zd/%zd; not retrying",
@@ -218,20 +211,37 @@ V1L_Flush(const struct worker *wrk)
 				break;
 			}
 
+			i = writev(*v1l->wfd, v1l->iov, v1l->niov);
+			if (i > 0)
+				v1l->cnt += i;
+
+			if (i == v1l->liov)
+				break;
+
+			/* we hit a timeout, and some data may have been sent:
+			 * Remove sent data from start of I/O vector, then retry
+			 *
+			 * XXX: Add a "minimum sent data per timeout counter to
+			 * prevent slowloris attacks
+			 */
+
 			VSLb(v1l->vsl, SLT_Debug,
 			    "Hit idle send timeout, wrote = %zd/%zd; retrying",
 			    i, v1l->liov);
 
-			v1l_prune(v1l, i);
-			i = writev(*v1l->wfd, v1l->iov, v1l->niov);
 			if (i > 0)
-				v1l->cnt += i;
-		}
+				v1l_prune(v1l, i);
+		} while (i > 0 || errno == EWOULDBLOCK);
+
 		if (i <= 0) {
-			v1l->werr++;
 			VSLb(v1l->vsl, SLT_Debug,
 			    "Write error, retval = %zd, len = %zd, errno = %s",
 			    i, v1l->liov, vstrerror(errno));
+			AZ(v1l->werr);
+			if (errno == EPIPE)
+				v1l->werr = SC_REM_CLOSE;
+			else
+				v1l->werr = SC_TX_ERROR;
 		}
 	}
 	v1l->liov = 0;
@@ -278,7 +288,7 @@ V1L_Chunked(const struct worker *wrk)
 	assert(v1l->ciov == v1l->siov);
 	assert(v1l->siov >= 3);
 	/*
-	 * If there are not space for chunked header, a chunk of data and
+	 * If there is no space for chunked header, a chunk of data and
 	 * a chunk tail, we might as well flush right away.
 	 */
 	if (v1l->niov + 3 >= v1l->siov)
