@@ -4,6 +4,8 @@
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
  *
+ * SPDX-License-Identifier: BSD-2-Clause
+ *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
  * are met:
@@ -86,11 +88,14 @@ h2_fini(struct req *req, void **priv)
 	struct h2_req *r2;
 
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
-	CAST_OBJ_NOTNULL(r2, *priv, H2_REQ_MAGIC);
+	TAKE_OBJ_NOTNULL(r2, priv, H2_REQ_MAGIC);
+
+	if (r2->error)
+		return (0);
+
 	H2_Send_Get(req->wrk, r2->h2sess, r2);
-	H2_Send(req->wrk, r2, H2_F_DATA, H2FF_DATA_END_STREAM, 0, "");
+	H2_Send(req->wrk, r2, H2_F_DATA, H2FF_DATA_END_STREAM, 0, "", NULL);
 	H2_Send_Rel(r2->h2sess, r2);
-	*priv = NULL;
 	return (0);
 }
 
@@ -106,10 +111,12 @@ h2_bytes(struct req *req, enum vdp_action act, void **priv,
 
 	if ((r2->h2sess->error || r2->error))
 		return (-1);
+	if (len == 0)
+		return (0);
 	H2_Send_Get(req->wrk, r2->h2sess, r2);
-	H2_Send(req->wrk, r2, H2_F_DATA, H2FF_NONE, len, ptr);
+	H2_Send(req->wrk, r2, H2_F_DATA, H2FF_NONE, len, ptr,
+	    &req->acct.resp_bodybytes);
 	H2_Send_Rel(r2->h2sess, r2);
-	req->acct.resp_bodybytes += len;
 	return (0);
 }
 
@@ -174,7 +181,7 @@ h2_minimal_response(struct req *req, uint16_t status)
 	    H2_F_HEADERS,
 	    H2FF_HEADERS_END_HEADERS |
 		(status < 200 ? 0 : H2FF_HEADERS_END_STREAM),
-	    l, buf);
+	    l, buf, NULL);
 	H2_Send_Rel(r2->h2sess, r2);
 	return (0);
 }
@@ -214,7 +221,7 @@ static const uint8_t h2_500_resp[] = {
 	0x1f, 0x27, 0x07, 'V', 'a', 'r', 'n', 'i', 's', 'h',
 };
 
-static int
+static void
 h2_build_headers(struct vsb *resp, struct req *req)
 {
 	unsigned u, l;
@@ -225,13 +232,8 @@ h2_build_headers(struct vsb *resp, struct req *req)
 	uint8_t buf[6];
 	ssize_t sz, sz1;
 
-	l = WS_Reserve(req->ws, 0);
-	if (l < 10)
-		return (-1);
-
-	AN(VSB_new(resp, req->ws->f, l, VSB_FIXEDLEN));
-
-	l = h2_status(buf, req->resp->status);
+	assert(req->resp->status % 1000 >= 100);
+	l = h2_status(buf, req->resp->status % 1000);
 	VSB_bcat(resp, buf, l);
 
 	hp = req->resp;
@@ -272,17 +274,17 @@ h2_build_headers(struct vsb *resp, struct req *req)
 		h2_enc_len(resp, 7, sz, 0);
 		VSB_bcat(resp, r, sz);
 	}
-	return (VSB_finish(resp));
 }
 
 void v_matchproto_(vtr_deliver_f)
 h2_deliver(struct req *req, struct boc *boc, int sendbody)
 {
-	ssize_t sz;
+	size_t sz;
 	const char *r;
 	struct sess *sp;
 	struct h2_req *r2;
-	struct vsb resp;
+	struct vsb resp[1];
+	uintptr_t ss;
 
 	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
 	CHECK_OBJ_ORNULL(boc, BOC_MAGIC);
@@ -293,9 +295,13 @@ h2_deliver(struct req *req, struct boc *boc, int sendbody)
 
 	VSLb(req->vsl, SLT_RespProtocol, "HTTP/2.0");
 
-	if (h2_build_headers(&resp, req)) {
-		// We ran out of workspace, return minimal 500
-		WS_MarkOverflow(req->ws);
+	ss = WS_Snapshot(req->ws);
+
+	WS_VSB_new(resp, req->ws);
+	h2_build_headers(resp, req);
+	r = WS_VSB_finish(resp, req->ws, &sz);
+
+	if (r == NULL) {
 		VSLb(req->vsl, SLT_Error, "workspace_client overflow");
 		VSLb(req->vsl, SLT_RespStatus, "500");
 		VSLb(req->vsl, SLT_RespReason, "Internal Server Error");
@@ -304,9 +310,6 @@ h2_deliver(struct req *req, struct boc *boc, int sendbody)
 		r = (const char*)h2_500_resp;
 		sz = sizeof h2_500_resp;
 		sendbody = 0;
-	} else {
-		sz = VSB_len(&resp);
-		r = req->ws->f;
 	}
 
 	AZ(req->wrk->v1l);
@@ -316,11 +319,10 @@ h2_deliver(struct req *req, struct boc *boc, int sendbody)
 	H2_Send_Get(req->wrk, r2->h2sess, r2);
 	H2_Send(req->wrk, r2, H2_F_HEADERS,
 	    (sendbody ? 0 : H2FF_HEADERS_END_STREAM) | H2FF_HEADERS_END_HEADERS,
-	    sz, r);
+	    sz, r, &req->acct.resp_hdrbytes);
 	H2_Send_Rel(r2->h2sess, r2);
-	req->acct.resp_hdrbytes += sz;
 
-	WS_Release(req->ws, 0);
+	WS_Reset(req->ws, ss);
 
 	/* XXX someone into H2 please add appropriate error handling */
 	if (sendbody) {
